"""Python logger settings."""

import os
import logging

# define custom logging levels
logging.DEBUG2 = 3
logging.addLevelName(logging.DEBUG2, "DEBUG-VERBOSE")
logging.__all__ += ["DEBUG2"]


class CustomFormatter(logging.Formatter):
    """Custom formatting for logger."""

    yellow = "\x1b[33;20m"
    green = "\x1b[32;20m"
    cyan = "\x1b[36;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    pre = "[%(asctime)s-%(name)s-%(levelname)s]"
    post = "(%(filename)s:%(funcName)s:%(lineno)d)"

    # Example [TIME:LEVEL:NAME] Message [FILE:FUNC:LINE]
    FORMATS = {
        logging.DEBUG: f"{cyan}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.DEBUG2: f"{cyan}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.INFO: f"{green}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.WARNING: f"{yellow}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.ERROR: f"{red}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.CRITICAL: f"{bold_red}{pre}{reset} %(message)s {yellow}{post}{reset}",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class CustomLogger:
    """
    Custom logger.
    Attributes:
        name (str): Logger's name.
    Usage:
        Set the ``<NAME>_LOGLEVEL`` environment variable to control the logging level for the whole project.
        If not set, the default logging level is ``DEBUG`` (0 = no logging, 1 = info, 2 = debug, 3 = debug verbose).
    """

    log_levels = {0: logging.NOTSET, 1: logging.INFO, 2: logging.DEBUG, 3: logging.DEBUG2}

    def __init__(self, name: str = "PANDAS_PARALLEL_APPLY"):

        # initialize logger and set logging level
        self.logger = logging.getLogger(name)
        env_var = int(os.environ[f"{name}_LOGLEVEL"]) if f"{name}_LOGLEVEL" in os.environ else 2
        self.logger.setLevel(self.log_levels[env_var])
        self.logger.debug2 = lambda msg: self.logger.log(logging.DEBUG2, msg)

        # add custom formatter to logger
        handler = logging.StreamHandler()
        handler.setFormatter(CustomFormatter())
        self.logger.addHandler(handler)

    def get_logger(self):
        """Get logger."""
        return self.logger

    def set_level(self, level: int):
        """
        Change logging level: 0 = no logging, 1 = info, 2 = debug, 3 = debug verbose.
        """
        if 0 <= level <= 3:
            self.logger.setLevel(self.log_levels[level])
        else:
            raise ValueError("Unknown logging level.")


# global project logger
logger = CustomLogger().get_logger()



"""
Apply on df.groupby
Usage:
    standard => df.groupby([cols]).apply(f)
    parallel => DataFrameParallel(df, n_cores).groupby([cols]).apply(f)
             => GroupByPrallel(df.groupby[cols], n_cores).apply(f)
"""
from typing import Callable, List, Union
from functools import partial
from pandas.core.groupby.generic import DataFrameGroupBy
from pandas.core.indexes.multi import MultiIndex
from multiprocessing import cpu_count
from tqdm import tqdm
import pandas as pd
from pathos.multiprocessing import ProcessingPool as Pool
import numpy as np
from .utils import get_default_n_cores
from .logger import logger


def _groupby_serial_func(data: List, func: Callable, pbar: bool = True) -> List:
    _range = tqdm(data) if pbar else data
    return [func(x) for x in _range]

def _get_multi_index_key(concat_res, key_data):
    new_key_data = []
    for i in range(len(concat_res)):
        # An awesome (KEY_1, 1), (KEY_2, 2), ..., (KEY_N, K) kind of deal
        n_keys = len(concat_res[i])
        multi_key = n_keys * [key_data[i]]
        counter_key = concat_res[i].index
        new_key = tuple(zip(multi_key, counter_key))
        new_key_data.extend(new_key)
    key_data = MultiIndex.from_tuples(new_key_data)
    return key_data

def chunk_df(df_grouped: DataFrameGroupBy, n_cores: int) -> List[List]:
    chunk_size = len(df_grouped) // n_cores + (len(df_grouped) % n_cores != 0)
    key_data = []
    chunked_data = []
    current_chunk = []
    for item in iter(df_grouped):
        key_data.append(item[0])
        current_chunk.append(item[1])
        if len(current_chunk) == chunk_size:
            chunked_data.append(current_chunk)
            current_chunk = []
    if 0 < len(current_chunk) < chunk_size:
        chunked_data.append(current_chunk)
    assert n_cores == len(chunked_data)
    return key_data, chunked_data

def apply_on_groupby_parallel(df_grouped: DataFrameGroupBy, func: Callable, n_cores: int, pbar: bool = True,
                              keep_original_indexes: bool = False) -> Union[pd.DataFrame, pd.Series]:
    if n_cores == 0:
        logger.info("n_cores is set to 0, returning serial apply.")
        return df_grouped.apply(func)

    n_cores = min(cpu_count(), len(df_grouped), n_cores)
    key_data, chunked_data = chunk_df(df_grouped, n_cores)

    # Run the multi-process job
    pool = Pool(n_cores)
    pool_res = pool.map(partial(_groupby_serial_func, func=func, pbar=pbar), chunked_data)

    # Concatenate the result to preserve the original result of a regular groupby pandas code.
    concat_res = []
    for i in range(len(pool_res)):
        concat_res.extend(pool_res[i])

    if isinstance(concat_res[0], pd.Series):
        res = pd.DataFrame(concat_res)
    elif isinstance(concat_res[0], pd.DataFrame):
        res = pd.concat(concat_res)
        # SLOWS DOWN BY A LOT, LIKE 5x SERIAL TIME. TO INVESTIGATE
        if keep_original_indexes:
            key_data = _get_multi_index_key(concat_res, key_data)
        else:
            key_data = np.arange(len(res))
    else:
        res = pd.Series(concat_res)

    # Fix the index
    res.index = key_data
    res.index.name = df_grouped.keys
    return res


class GroupByParallel:
    def __init__(self, df_grouped: DataFrameGroupBy, n_cores: int = None, pbar: bool = True,
                 keep_original_indexes: bool = False):
        if n_cores is None:
            n_cores = get_default_n_cores()
        self.df_grouped = df_grouped
        self.n_cores = n_cores
        self.pbar = pbar
        self.keep_original_indexes = keep_original_indexes
    
    def apply(self, func: Callable):
        return apply_on_groupby_parallel(self.df_grouped, func, self.n_cores, self.pbar, self.keep_original_indexes)



import pandas as pd
from pandas_parallel_apply import GroupByParallel
import numpy as np

from pandas_parallel_apply.data_frame_parallel import DataFrameParallel

data_static = {
    "A": [1,1,2,3,5,3],
    "B": ["hello", "darkness", "hi", "dafuq", "who", "asdf"],
    1: [(1,2), (3,4), "yolo", 1, 2, 3]
}

N = 10000000
data_random = {"A": np.random.randint(0, 10, size=(N, )), "B": np.arange(N),
    "C": [chr(ord("A") + np.random.randint(0, 26)) for _ in range(N)] }

def test_one_col_return_series():
    df = pd.DataFrame(data_random)
    Y = df.groupby("A").apply(len)
    K = GroupByParallel(df.groupby("A"), pbar=False).apply(len)
    P = DataFrameParallel(df, pbar=False).groupby("A").apply(len)

    assert np.allclose(Y.values, K.values)
    assert np.allclose(Y.values, P.values)

def test_one_col_return_df():
    df = pd.DataFrame(data_random)
    f = lambda df: df.iloc[0]
    Y = df.groupby("A").apply(f)
    K = GroupByParallel(df.groupby("A"), pbar=False).apply(f)
    P = DataFrameParallel(df, pbar=False).groupby("A").apply(f)
    assert (Y != K).sum().sum() == 0
    assert (Y != P).sum().sum() == 0

def run_one_col_return_tricky_df_no_index(df):
    def f(df):
        np.random.seed(42)
        N = np.random.randint(1, len(df) + 1)
        return df.iloc[0 : N]

    df = pd.DataFrame(data_random)
    Y = df.groupby("A").apply(f)
    K = GroupByParallel(df.groupby("A"), pbar=False, keep_original_indexes=False).apply(f)
    P = DataFrameParallel(df, pbar=False).groupby("A").apply(f)
    assert (Y.values != K.values).sum() == 0
    assert (Y.values != P.values).sum() == 0

def run_one_col_return_tricky_df_plus_index(df):
    def f(df):
        np.random.seed(42)
        N = np.random.randint(1, len(df) + 1)
        return df.iloc[0 : N]

    df = pd.DataFrame(data_random)
    Y = df.groupby("A").apply(f)
    K = GroupByParallel(df.groupby("A"), pbar=False, keep_original_indexes=True).apply(f)
    assert (Y != K).sum().sum() == 0



import pandas as pd
from pandas_parallel_apply import DataFrameParallel
from datetime import datetime
import numpy as np

data = {
    "A": [1,2,3],
    "B": ["hello", "darkness", "hi"],
    1: [(1,2), (3,4), "yolo"]
}

data = np.random.randn(100000, 3)
df = pd.DataFrame(data, columns=["A", "B", 1])

def f(x):
    return [x["A"] + 99, x["B"], x[1]]

def test_apply_df_1():
    asdf = df.apply(f, axis=1)
    dfp = DataFrameParallel(df, 4, pbar=False)
    asdf2 = dfp.apply(f, axis=1)
    assert np.allclose(np.concatenate(asdf), np.concatenate(asdf2))



"""
Wrapper on top of pd.DataFrame with some parallel operations
Apply on column
Usage:
    standard => df[col_name].apply(f)
    parallel => DataFrameParallel(df, n_cores)[col_name].apply(f)

Apply on row
Usage:
    standard => df.apply(f, axis=1)
    parallel => DataFrameParallel(df, n_cores).apply(f, axis=1)
"""
import pandas as pd

from .series_parallel import SeriesParallel
from .groupby_parallel import GroupByParallel
from typing import Callable
from .utils import parallelize_dataframe, get_default_n_cores
from functools import partial


def apply_on_df(df: pd.DataFrame, f: Callable, pbar: bool = True) -> pd.Series:
    """Apply a function on each row (all possible columns), returning a series"""
    if pbar:
        return df.progress_apply(f, axis=1)
    else:
        return df.apply(f, axis=1)

class DataFrameParallel:
    def __init__(self, df: pd.DataFrame, n_cores: int = None, pbar: bool = True):
        if n_cores is None:
            n_cores = get_default_n_cores()

        self.df = df
        self.n_cores = n_cores
        self.pbar = pbar

    def apply(self, func, axis, raw: bool = False, result_type = None, args=(), **kwargs):
        assert axis == 1, "Only axis=1 is supported in parallel df apply"
        return parallelize_dataframe(self.df, partial(apply_on_df, f=func, pbar=self.pbar), self.n_cores)

    def groupby(self, *args, **kwargs):
        return GroupByParallel(self.df.groupby(*args, **kwargs), self.n_cores, self.pbar)

    def __getitem__(self, x):
        return SeriesParallel(self.df[x], self.n_cores, self.pbar)

    def __str__(self) -> str:
        f_str = f"[Parallel DataFrame - {self.n_cores} crores]\n" + self.df.__str__()
        return f_str



"""
Apply on series
Usage:
    standard => series.apply(f)
    parallel => SeriesParallel(series, n_cores).apply(f)
             => SeriesParallel(df[col], n_cores).apply(f)
"""
from typing import Callable
import pandas as pd
from functools import partial

from .utils import parallelize_dataframe, get_default_n_cores


def apply_on_series(series: pd.Series, f: Callable, pbar: bool = True) -> pd.Series:
    if pbar:
        return series.progress_apply(f)
    else:
        return series.apply(f)


class SeriesParallel:
    def __init__(self, series: pd.Series, n_cores: int = None, pbar: bool = True):
        if n_cores is None:
            n_cores = get_default_n_cores()
        self.series = series
        self.n_cores = n_cores
        self.pbar = pbar
    
    def apply(self, func: Callable) -> pd.Series:
        return parallelize_dataframe(self.series, partial(apply_on_series, f=func, pbar=self.pbar), self.n_cores)



"""General module for parallelizing a dataframe apply function on a column (series) or entire row"""

import logging
from typing import Callable, Union
import numpy as np
import pandas as pd
from tqdm import tqdm
from multiprocessing import cpu_count
from pathos.multiprocessing import ProcessingPool as Pool
import os
from .logger import logger
tqdm.pandas()


def parallelize_dataframe(df: Union[pd.DataFrame, pd.Series], func: Callable, n_cores: int) -> pd.DataFrame:
    """Function used to split a dataframe in n sub dataframes, based on the number of cores we want to use."""
    if n_cores == 0:
        return func(df)
    n_cores = min(cpu_count(), len(df), n_cores)
    logging.debug(f"Parallelizing apply on df (rows: {len(df)}) with {n_cores} cores")
    df_split = np.array_split(df, n_cores)
    pool = Pool(n_cores)
    pool_res = pool.map(func, df_split)
    df = pd.concat(pool_res)
    return df

def get_default_n_cores() -> int:
    """Gets the default N_CORES if set by the PANDAS_PARALLEL_APPLY_N_CORES env variable, else set to 0 (serial)"""
    if "PANDAS_PARALLEL_APPLY_N_CORES" in os.environ:
        n_cores = int(os.environ["PANDAS_PARALLEL_APPLY_N_CORES"])
        logger.debug(f"N cores was set from the 'PANDAS_PARALLEL_APPLY_N_CORES' env variable to {n_cores}")
    else:
        n_cores = 0
        logger.debug(f"No 'PANDAS_PARALLEL_APPLY_N_CORES' env variable. Defaulting n_cores to 0 (serial)")
    assert isinstance(n_cores, int) and n_cores >= 0
    return n_cores



import pandas as pd
from pandas_parallel_apply import DataFrameParallel, SeriesParallel
import numpy as np

data = np.random.randn(1000000, 3)
df = pd.DataFrame(data, columns=["A", "B", 1])

def f(x):
    return x + 99

def test_apply_df_col_1():
    asdf = df["A"].apply(f)

    dfp = DataFrameParallel(df, pbar=False)
    asdf2 = dfp["A"].apply(f)
    assert np.allclose(asdf, asdf2)

    asdf4 = SeriesParallel(df["A"], pbar=False).apply(f)
    assert np.allclose(asdf, asdf4)
